/*
           _        _    _            _
     /\   (_)      | |  | |          | |
    /  \   _ _ __  | |__| | ___   ___| | _____ _   _
   / /\ \ | | '__| |  __  |/ _ \ / __| |/ / _ \ | | |
  / ____ \| | |    | |  | | (_) | (__|   <  __/ |_| |
 /_/    \_\_|_|    |_|  |_|\___/ \___|_|\_\___|\__, |
                                                __/ |
                                               |___/
by David Mason
*/

#include <LiquidCrystal.h>

LiquidCrystal lcd(A5, A4, A0, A1, A3, A2);   // Arduino Uno

//////////////////////
// Constants
//////////////////////
// Button Inputs
const int switchStart = 8;
const int switchGoal1 = 9;
const int switchGoal2 = 10;

// MOSFET Outputs
int ledState = LOW;
const int mosfetLED1 = 5;
const int mosfetLED2 = 6;
const int mosfetFan = 7;

#define HIGH_SCORE 5
#define MAX_PLAY_TIME 180000 //90000 // 3 mins
#define GOAL_CELEBRATION_TIME 3000 // 3 seconds
#define LED_INTERVAL 200 // 5th of a second
#define GAMEOVER_DELAY 4000

// Game states
enum {
  GS_ReadyToPlay = 0,
  GS_Playing,
  GS_Goal,
  GS_GameOver
} GameStates;


//////////////////////
// Variables
//////////////////////

// Globals stored across games
int p1Games = 0;
int p2Games = 0;
int led1Brightness = 0;
int led2Brightness = 0;
int ledFadeDirection = 1;

// Globals reset per game
int p1Score = 0;
int p2Score = 0;
unsigned long playStartTime = 0;
unsigned long ledInterval = 0;
unsigned long ledIntervalPrev = 0;
int gameState = GS_ReadyToPlay;

// Button States
int bsStart;
int bsStartPrev;
int bsGoal1;
int bsGoal1Prev;
int bsGoal2;
int bsGoal2Prev;
/*
  _      _____ _____    __  __
 | |    / ____|  __ \  |  \/  |
 | |   | |    | |  | | | \  / | ___  ___ ___  __ _  __ _  ___  ___
 | |   | |    | |  | | | |\/| |/ _ \/ __/ __|/ _` |/ _` |/ _ \/ __|
 | |___| |____| |__| | | |  | |  __/\__ \__ \ (_| | (_| |  __/\__ \
 |______\_____|_____/  |_|  |_|\___||___/___/\__,_|\__, |\___||___/
                                                    __/ |
                                                   |___/
*/


void displayWelcomeMessage(void)
{
  lcd.clear();
  lcd.print ("Press to Play!");
  lcd.setCursor(0, 1);
  lcd.print("Games P1:");
  lcd.print(p1Games);
  lcd.print(" P2:");
  lcd.print(p2Games);
  lcd.setCursor(0, 0);
}

void inGameDisplay(void)
{
  lcd.print ("P1: ");
  lcd.print (p1Score);
  lcd.print ("   P2: ");
  lcd.print (p2Score);
  lcd.setCursor (0, 1);
  lcd.print ("Time left: ");
  int counter = (((playStartTime + MAX_PLAY_TIME) - millis()) / 1000);
  lcd.print (counter);
  lcd.print (" ");
  lcd.setCursor(0, 0);
}

void P1WinsDisplay(void)
{
  p1Games++;
  lcd.clear();
  lcd.print ("Player One");
  lcd.setCursor(0, 1);
  lcd.print("wins!");
  lcd.setCursor(0, 0);
}

void P2WinsDisplay(void)
{
  p2Games++;
  lcd.clear();
  lcd.print ("Player Two");
  lcd.setCursor(0, 1);
  lcd.print("wins!");
  lcd.setCursor(0, 0);
}

void P1Scored(void)
{
  lcd.clear();
  lcd.print ("Player One");
  lcd.setCursor(0, 1);
  lcd.print("has scored!");
  lcd.setCursor(0, 0);
}

void P2Scored(void)
{
  lcd.clear();
  lcd.print ("Player Two");
  lcd.setCursor(0, 1);
  lcd.print("has scored!");
  lcd.setCursor(0, 0);
}

void Draw(void)
{
  lcd.clear();
  lcd.print ("It's a draw!");
  lcd.setCursor(0, 1);
  lcd.print("Both scored ");
  lcd.print(p1Score);
  lcd.setCursor(0, 0);
}

// Check what to display and move to game over state
void DisplayGameOverState(void)
{
  digitalWrite (mosfetFan, LOW);
  // Decide what to display
  if(p1Score == p2Score) {
    Draw();
  } else if(p1Score > p2Score) {
    P1WinsDisplay();
  } else {
    P2WinsDisplay();
  }
  gameState = GS_GameOver;
}


void setup() {

  // Setup pins for LCD
  pinMode(A0, OUTPUT);
  pinMode(A1, OUTPUT);
  pinMode(A2, OUTPUT);
  pinMode(A3, OUTPUT);
  pinMode(A4, OUTPUT);
  pinMode(A5, OUTPUT);
  lcd.begin(16, 2);

  // Start serial interface for debug
  Serial.begin(9600);

  // Setup switches (all require internal pullup)
  pinMode (switchStart, INPUT_PULLUP);
  pinMode (switchGoal1, INPUT_PULLUP);
  pinMode (switchGoal2, INPUT_PULLUP);
  bsStartPrev = digitalRead(switchStart);
  bsGoal1Prev = digitalRead(switchGoal1);
  bsGoal2Prev = digitalRead(switchGoal2);

  // Setup MOSFET Outputs
  digitalWrite (mosfetLED1, LOW);
  digitalWrite (mosfetLED2, LOW);
  digitalWrite (mosfetFan, LOW);
  pinMode(mosfetLED1, OUTPUT);
  pinMode(mosfetLED2, OUTPUT);
  pinMode(mosfetFan, OUTPUT);

  // Display welcome message
  displayWelcomeMessage();

  // Initialise the game state
  gameState = GS_ReadyToPlay;
}

void loop() {
  boolean startPressed = false, goal1Pressed = false, goal2Pressed = false;
  boolean startReleased = false, goal1Released = false, goal2Released = false;
  boolean p1Win = false, p2Win = false;

  // Read switches
  bsStart = digitalRead(switchStart);
  bsGoal1 = digitalRead(switchGoal1);
  bsGoal2 = digitalRead(switchGoal2);

  // Decide if a switch has changed
  if (bsStart != bsStartPrev) {
    if (bsStart == 0) startPressed = true;
    else startReleased = true;
  }
  if (bsGoal1 != bsGoal1Prev) {
    if (bsGoal1 == 0) goal1Pressed = true;
    else goal1Released = true;
  }
  if (bsGoal2 != bsGoal2Prev) {
    if (bsGoal2 == 0) goal2Pressed = true;
    else goal2Released = true;
  }


  // Set LED arrays to current fade value
  analogWrite(mosfetLED1, led1Brightness);
  analogWrite(mosfetLED2, led2Brightness);


  // Check for game start
  switch (gameState)
  {
    default:
    case GS_ReadyToPlay:
      // Fade LEDs
      led1Brightness += ledFadeDirection;
      led2Brightness += ledFadeDirection;
      
      if(led1Brightness >= 255) ledFadeDirection = -2;
      if(led1Brightness <= 1) ledFadeDirection = 2;
      

      // Check for Start button
      if (startPressed) {

        // Move to 'Playing' game state
        gameState = GS_Playing;

        // Reset scores
        p1Score = 0; p1Win = false;
        p2Score = 0; p2Win = false;

        // Record time at which we started the game
        playStartTime = millis();
        //digitalWrite (mosfetFan, HIGH);

        // Change message on LCD
        lcd.clear();

        // Set LED Array to value for play
        led1Brightness = 128; // 50% brightness
        led2Brightness = 128;
      }
      break;

    case GS_Playing:
      // Check if someon has won
      if ((p1Score >= HIGH_SCORE) || (p2Score >= HIGH_SCORE)) {
        DisplayGameOverState();
        break;
      }
 
      // Display a message with current scores and game time remaining
      inGameDisplay();
      // Check goal switches and update scores

      if (goal1Pressed)
      {
        p2Score++;
        P2Scored();
        gameState = GS_Goal;
      }

      if (goal2Pressed)
      {
        p1Score++;
        P1Scored();
        gameState = GS_Goal;
      }

      if (startPressed)
      {
        // Abort game
        gameState = GS_ReadyToPlay;
        // Display welcome message
        displayWelcomeMessage();
      }
      if (millis() > (playStartTime + MAX_PLAY_TIME)) {
        DisplayGameOverState();
        break;

        // Game timeout
        //gameState = GS_ReadyToPlay;
        // Display welcome message
        //displayWelcomeMessage();
      }
      break;

    case GS_Goal:
      {
        unsigned long goalCelebrationStart = 0;      // To allow for goal celebration

        // Flash LED's for 3 seconds

        goalCelebrationStart = millis();

        // Do not accept goals for 3 seconds whilst we flash LEDS
        while (millis() <= goalCelebrationStart + GOAL_CELEBRATION_TIME)
        {
          ledInterval = millis();
          if (ledInterval - ledIntervalPrev >= LED_INTERVAL) {
            ledIntervalPrev = ledInterval;
            if (ledState == LOW)
            {
              ledState = HIGH;
            }
            else
            {
              ledState = LOW;
            }
            digitalWrite (mosfetLED1, ledState);
            digitalWrite (mosfetLED2, ledState);
          }
        }

        playStartTime += GOAL_CELEBRATION_TIME; // add a few seconds to length of the game
        gameState = GS_Playing;
      }
      break;

    case GS_GameOver:
      {
        unsigned long gameOverStartTime = millis();
        
        while(millis() - gameOverStartTime <= GAMEOVER_DELAY) {
          digitalWrite (mosfetFan, LOW);
        }
      }
      
      // Display welcome message
      displayWelcomeMessage();

      gameState = GS_ReadyToPlay;
      break;
  }

  // Update our previous states ready for next loop iteration
  bsStartPrev = bsStart;
  bsGoal1Prev = bsGoal1;
  bsGoal2Prev = bsGoal2;
  delay(10);
}
